----------------------------------------------------------------------------------
--                                 _             _
--                                | |_  ___ _ __(_)__ _
--                                | ' \/ -_) '_ \ / _` |
--                                |_||_\___| .__/_\__,_|
--                                         |_|
--
----------------------------------------------------------------------------------
--
-- Company: hepia
-- Author: Joachim Schmidt <joachim.schmidt@hesge.ch
--
-- Module Name: clock_module - rtl
-- Target Device: SCALP xc7z015clg485-2
-- Tool version: 2020.2
-- Description: Aurora clock module
--
-- Last update: 2020-09-28
--
---------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library UNISIM;
use UNISIM.VCOMPONENTS.all;

entity clock_module is

    generic (
        C_NUM_PLL : integer := 1);
    port (
        InitClkxCI      : in  std_logic;
        InitClkxCO      : out std_logic;
        GTClkxCI        : in  std_logic_vector((C_NUM_PLL - 1) downto 0);
        GTClkLockedxDI  : in  std_logic_vector((C_NUM_PLL - 1) downto 0);
        UserClkxCO      : out std_logic_vector((C_NUM_PLL - 1) downto 0);
        SyncClkxCO      : out std_logic_vector((C_NUM_PLL - 1) downto 0);
        PLLNotLockedxDO : out std_logic_vector((C_NUM_PLL - 1) downto 0));

end entity clock_module;

architecture rtl of clock_module is

    -- PLL Constants
    ---------------------------------------------------------------------------
    -- 2.5 Gbps
    -- constant C_MULT        : real    := 9.0;
    -- constant C_DIVIDE      : integer := 1;
    -- constant C_CLK_PERIOD  : real    := 8.000;
    -- constant C_OUT0_DIVIDE : real    := 18.0;
    -- constant C_OUT1_DIVIDE : integer := 9;
    -- constant C_OUT2_DIVIDE : integer := 18;
    -- constant C_OUT3_DIVIDE : integer := 9;
    -- 5 Gbps
    constant C_MULT        : real    := 8.0;
    constant C_DIVIDE      : integer := 2;
    constant C_CLK_PERIOD  : real    := 4.000;
    constant C_OUT0_DIVIDE : real    := 8.0;
    constant C_OUT1_DIVIDE : integer := 4;
    constant C_OUT2_DIVIDE : integer := 8;
    constant C_OUT3_DIVIDE : integer := 4;

    -- Components
    ---------------------------------------------------------------------------
    -- Clock Buffer
    component BUFG
        port (
            O : out std_ulogic;
            I : in  std_ulogic);
    end component;

    -- Advance MMCME2 (PLL)
    component MMCME2_ADV
        generic (
            BANDWIDTH            : string  := "OPTIMIZED";
            CLKFBOUT_MULT_F      : real    := 5.000;
            CLKFBOUT_PHASE       : real    := 0.000;
            CLKFBOUT_USE_FINE_PS : boolean := false;
            CLKIN1_PERIOD        : real    := 0.000;
            CLKIN2_PERIOD        : real    := 0.000;
            CLKOUT0_DIVIDE_F     : real    := 1.000;
            CLKOUT0_DUTY_CYCLE   : real    := 0.500;
            CLKOUT0_PHASE        : real    := 0.000;
            CLKOUT0_USE_FINE_PS  : boolean := false;
            CLKOUT1_DIVIDE       : integer := 1;
            CLKOUT1_DUTY_CYCLE   : real    := 0.500;
            CLKOUT1_PHASE        : real    := 0.000;
            CLKOUT1_USE_FINE_PS  : boolean := false;
            CLKOUT2_DIVIDE       : integer := 1;
            CLKOUT2_DUTY_CYCLE   : real    := 0.500;
            CLKOUT2_PHASE        : real    := 0.000;
            CLKOUT2_USE_FINE_PS  : boolean := false;
            CLKOUT3_DIVIDE       : integer := 1;
            CLKOUT3_DUTY_CYCLE   : real    := 0.500;
            CLKOUT3_PHASE        : real    := 0.000;
            CLKOUT3_USE_FINE_PS  : boolean := false;
            CLKOUT4_CASCADE      : boolean := false;
            CLKOUT4_DIVIDE       : integer := 1;
            CLKOUT4_DUTY_CYCLE   : real    := 0.500;
            CLKOUT4_PHASE        : real    := 0.000;
            CLKOUT4_USE_FINE_PS  : boolean := false;
            CLKOUT5_DIVIDE       : integer := 1;
            CLKOUT5_DUTY_CYCLE   : real    := 0.500;
            CLKOUT5_PHASE        : real    := 0.000;
            CLKOUT5_USE_FINE_PS  : boolean := false;
            CLKOUT6_DIVIDE       : integer := 1;
            CLKOUT6_DUTY_CYCLE   : real    := 0.500;
            CLKOUT6_PHASE        : real    := 0.000;
            CLKOUT6_USE_FINE_PS  : boolean := false;
            COMPENSATION         : string  := "ZHOLD";
            DIVCLK_DIVIDE        : integer := 1;
            REF_JITTER1          : real    := 0.0;
            REF_JITTER2          : real    := 0.0;
            SS_EN                : string  := "FALSE";
            SS_MODE              : string  := "CENTER_HIGH";
            SS_MOD_PERIOD        : integer := 10000;
            STARTUP_WAIT         : boolean := false);
        port (
            CLKFBOUT     : out std_ulogic := '0';
            CLKFBOUTB    : out std_ulogic := '0';
            CLKFBSTOPPED : out std_ulogic := '0';
            CLKINSTOPPED : out std_ulogic := '0';
            CLKOUT0      : out std_ulogic := '0';
            CLKOUT0B     : out std_ulogic := '0';
            CLKOUT1      : out std_ulogic := '0';
            CLKOUT1B     : out std_ulogic := '0';
            CLKOUT2      : out std_ulogic := '0';
            CLKOUT2B     : out std_ulogic := '0';
            CLKOUT3      : out std_ulogic := '0';
            CLKOUT3B     : out std_ulogic := '0';
            CLKOUT4      : out std_ulogic := '0';
            CLKOUT5      : out std_ulogic := '0';
            CLKOUT6      : out std_ulogic := '0';
            DO           : out std_logic_vector (15 downto 0);
            DRDY         : out std_ulogic := '0';
            LOCKED       : out std_ulogic := '0';
            PSDONE       : out std_ulogic := '0';
            CLKFBIN      : in  std_ulogic;
            CLKIN1       : in  std_ulogic;
            CLKIN2       : in  std_ulogic;
            CLKINSEL     : in  std_ulogic;
            DADDR        : in  std_logic_vector(6 downto 0);
            DCLK         : in  std_ulogic;
            DEN          : in  std_ulogic;
            DI           : in  std_logic_vector(15 downto 0);
            DWE          : in  std_ulogic;
            PSCLK        : in  std_ulogic;
            PSEN         : in  std_ulogic;
            PSINCDEC     : in  std_ulogic;
            PWRDWN       : in  std_ulogic;
            RST          : in  std_ulogic);
    end component;

    -- Signals
    ---------------------------------------------------------------------------
    signal LockedWxD            : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ResetxRN             : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    -- Input clock buffering
    signal ClkIn1xC             : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    -- Output clock buffering
    signal ClkFBOutxC           : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkFBOutBUnusedxC    : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut0xC            : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut0BUnusedxC     : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut1xC            : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut1BUnusedxC     : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut2xC            : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut2BUnusedxC     : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut3xC            : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut3BUnusedxC     : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut4UnusedxC      : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut5UnusedxC      : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkOut6UnusedxC      : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    -- Dynamic programming unused signals
    signal DoUnusedxD           : std_logic_vector(((C_NUM_PLL * 15) + (C_NUM_PLL - 1)) downto 0) := (others => '0');
    signal DrdyUnusedxD         : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    -- Dynamic phase shift unused signals
    signal PSDoneUnusedxD       : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    -- Unused status signals
    signal ClkFBStoppedUnusedxD : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');
    signal ClkInStoppedUnusedxD : std_logic_vector((C_NUM_PLL - 1) downto 0)                      := (others => '0');

begin  -- architecture rtl    

    ClockModulexB : block is
    begin  -- block ClockModulex

        InitClkBufGxI : BUFG
            port map (
                I => InitClkxCI,
                O => InitClkxCO);

        ClockBufxG : for i in 0 to (C_NUM_PLL - 1) generate

            PLLResetxAS     : ResetxRN(i)        <= not GTClkLockedxDI(i);
            PLLNotLockedxAS : PLLNotLockedxDO(i) <= not LockedWxD(i);

            ClkIn1BufGxI : BUFG
                port map (
                    O => ClkIn1xC(i),
                    I => GTClkxCI(i));

            ClkOut0BufGxI : BUFG
                port map (
                    O => UserClkxCO(i),
                    I => ClkOut0xC(i));

            ClkOut1BufGxI : BUFG
                port map (
                    O => SyncClkxCO(i),
                    I => ClkOut1xC(i));

        end generate ClockBufxG;

        MMCME2AdvxG : for i in 0 to (C_NUM_PLL - 1) generate

            MMCME2AdvxI : MMCME2_ADV
                generic map (
                    BANDWIDTH            => "OPTIMIZED",
                    CLKOUT4_CASCADE      => false,
                    COMPENSATION         => "ZHOLD",
                    STARTUP_WAIT         => false,
                    DIVCLK_DIVIDE        => C_DIVIDE,
                    CLKFBOUT_MULT_F      => C_MULT,
                    CLKFBOUT_PHASE       => 0.000,
                    CLKFBOUT_USE_FINE_PS => false,
                    CLKOUT0_DIVIDE_F     => C_OUT0_DIVIDE,
                    CLKOUT0_PHASE        => 0.000,
                    CLKOUT0_DUTY_CYCLE   => 0.500,
                    CLKOUT0_USE_FINE_PS  => false,
                    CLKIN1_PERIOD        => C_CLK_PERIOD,
                    CLKOUT1_DIVIDE       => C_OUT1_DIVIDE,
                    CLKOUT1_PHASE        => 0.000,
                    CLKOUT1_DUTY_CYCLE   => 0.500,
                    CLKOUT1_USE_FINE_PS  => false,
                    CLKOUT2_DIVIDE       => C_OUT2_DIVIDE,
                    CLKOUT2_PHASE        => 0.000,
                    CLKOUT2_DUTY_CYCLE   => 0.500,
                    CLKOUT2_USE_FINE_PS  => false,
                    CLKOUT3_DIVIDE       => C_OUT3_DIVIDE,
                    CLKOUT3_PHASE        => 0.000,
                    CLKOUT3_DUTY_CYCLE   => 0.500,
                    CLKOUT3_USE_FINE_PS  => false,
                    REF_JITTER1          => 0.010)
                port map (
                    -- Output clocks
                    CLKFBOUT     => ClkFBOutxC(i),
                    CLKFBOUTB    => ClkFBOutBUnusedxC(i),
                    CLKOUT0      => ClkOut0xC(i),
                    CLKOUT0B     => ClkOut0BUnusedxC(i),
                    CLKOUT1      => ClkOut1xC(i),
                    CLKOUT1B     => ClkOut1BUnusedxC(i),
                    CLKOUT2      => ClkOut2xC(i),
                    CLKOUT2B     => ClkOut2BUnusedxC(i),
                    CLKOUT3      => ClkOut3xC(i),
                    CLKOUT3B     => ClkOut3BUnusedxC(i),
                    CLKOUT4      => ClkOut4UnusedxC(i),
                    CLKOUT5      => ClkOut5UnusedxC(i),
                    CLKOUT6      => ClkOut6UnusedxC(i),
                    -- Input clock control
                    CLKFBIN      => ClkFBOutxC(i),
                    CLKIN1       => ClkIn1xC(i),
                    CLKIN2       => '0',
                    -- Tied to always select the primary input clock
                    CLKINSEL     => '1',
                    -- Ports for dynamic reconfiguration
                    DADDR        => (others => '0'),
                    DCLK         => '0',
                    DEN          => '0',
                    DI           => (others => '0'),
                    DO           => DoUnusedxD((((i + 1) * 15) + i) downto ((i * 15) + i)),
                    DRDY         => DrdyUnusedxD(i),
                    DWE          => '0',
                    -- Ports for dynamic phase shift
                    PSCLK        => '0',
                    PSEN         => '0',
                    PSINCDEC     => '0',
                    PSDONE       => PSDoneUnusedxD(i),
                    -- Other control and status signals
                    LOCKED       => LockedWxD(i),
                    CLKINSTOPPED => ClkFBStoppedUnusedxD(i),
                    CLKFBSTOPPED => ClkInStoppedUnusedxD(i),
                    PWRDWN       => '0',
                    RST          => ResetxRN(i));

        end generate MMCME2AdvxG;

    end block ClockModulexB;

end architecture rtl;
