----------------------------------------------------------------------------------
--                                 _             _
--                                | |_  ___ _ __(_)__ _
--                                | ' \/ -_) '_ \ / _` |
--                                |_||_\___| .__/_\__,_|
--                                         |_|
--
----------------------------------------------------------------------------------
--
-- Company: hepia
-- Author: Joachim Schmidt <joachim.schmidt@hesge.ch>
--
-- Module Name: scalp_misc.vhd
-- Target Device: SCALP xc7z015clg485-2
-- Tool version: 2018.2
-- Description: Scalp Miscellanous
--
-- Last update: 2020-12-07
--
---------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.axi4_pkg.all;

package scalp_misc is

    constant C_SCALP_X_RANGE_VALUE             : integer range 0 to 255   := 255;
    constant C_SCALP_Y_RANGE_VALUE             : integer range 0 to 255   := 255;
    constant C_SCALP_Z_RANGE_VALUE             : integer range 0 to 255   := 255;
    constant C_SCALP_PACKET_TYPE_RANGE_VALUE   : integer range 0 to 255   := 255;
    constant C_SCALP_PACKET_LENGTH_RANGE_VALUE : integer range 0 to 65535 := 65535;
    constant C_SCALP_ADDR_SIZE                 : integer range 0 to 255   := 8;
    constant C_BYTE_SIZE                       : integer range 0 to 255   := 8;
    -- Scalp router interfaces id.
    constant C_NORTH_IF_ID                     : integer range 0 to 6     := 0;
    constant C_EAST_IF_ID                      : integer range 0 to 6     := 1;
    constant C_SOUTH_IF_ID                     : integer range 0 to 6     := 2;
    constant C_WEST_IF_ID                      : integer range 0 to 6     := 3;
    constant C_TOP_IF_ID                       : integer range 0 to 6     := 4;
    constant C_BOTTOM_IF_ID                    : integer range 0 to 6     := 5;
    constant C_LOCAL_IF_ID                     : integer range 0 to 6     := 6;
    -- Scalp payload sizes
    constant C_SCALP_PACKET_PAYLOAD_SIZE_1     : integer range 0 to 256   := 1;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_2     : integer range 0 to 256   := 2;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_4     : integer range 0 to 256   := 4;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_8     : integer range 0 to 256   := 8;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_16    : integer range 0 to 256   := 16;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_32    : integer range 0 to 256   := 32;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_64    : integer range 0 to 256   := 64;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_128   : integer range 0 to 256   := 128;
    constant C_SCALP_PACKET_PAYLOAD_SIZE_256   : integer range 0 to 256   := 256;
    -- Scalp Simulation
    constant C_SCALP_SIMULATION                : boolean;

    -- Scalp Booking Vector States
    type t_scalp_booking_vector_states is (E_SCALP_BOOKING_VECTOR_VALID_INDEX,
                                           E_SCALP_BOOKING_VECTOR_SCHEDULER_ACK);

    -- Scalp Tx Side States
    type t_scalp_tx_side_states is (E_SCALP_TX_IDLE,
                                    E_SCALP_TX_WAIT_BOOKING_VECTOR,
                                    E_SCALP_TX_POP_N_PUSH);

    -- Scalp Rx Side States
    type t_scalp_rx_side_states is (E_SCALP_RX_IDLE,
                                    E_SCALP_RX_ARBITRATE,
                                    E_SCALP_RX_WAIT_ARBITRATED,
                                    E_SCALP_RX_POP_N_PUSH,
                                    E_SCALP_RX_ACK);

    -- Scalp Packet States
    type t_scalp_packet_states is (E_SP_IDLE,
                                   E_SP_WAIT,
                                   E_SP_H0,
                                   E_SP_H1,
                                   E_SP_H2,
                                   E_SP_P,
                                   E_SP_P_LAST);

    type t_scalp_netaddr is record
        XxD : integer range 0 to C_SCALP_X_RANGE_VALUE;
        YxD : integer range 0 to C_SCALP_Y_RANGE_VALUE;
        ZxD : integer range 0 to C_SCALP_Z_RANGE_VALUE;
    end record t_scalp_netaddr;

    constant C_3D_MIN_SCALP_NETADDR : t_scalp_netaddr := (XxD => 0,
                                                          YxD => 0,
                                                          ZxD => 0);
    constant C_3D_MAX_SCALP_NETADDR : t_scalp_netaddr := (XxD => C_SCALP_X_RANGE_VALUE,
                                                          YxD => C_SCALP_Y_RANGE_VALUE,
                                                          ZxD => C_SCALP_Z_RANGE_VALUE);
    constant C_2D_MIN_SCALP_NETADDR : t_scalp_netaddr := (XxD => 0,
                                                          YxD => 0,
                                                          ZxD => 0);
    constant C_2D_MAX_SCALP_NETADDR : t_scalp_netaddr := (XxD => C_SCALP_X_RANGE_VALUE,
                                                          YxD => C_SCALP_Y_RANGE_VALUE,
                                                          ZxD => 0);

    type t_scalp_netaddr_vector is array (natural range <>) of t_scalp_netaddr;

    -- Scalp QoS Vector
    constant C_SCALP_QOS_SIZE : integer range 0 to 7 := 7;

    subtype t_scalp_qos is unsigned((C_SCALP_QOS_SIZE - 1) downto 0);
    type t_scalp_qos_vector is array (natural range <>) of t_scalp_qos;

    constant C_SCALP_NO_QOS : t_scalp_qos := (others => '0');

    -- Scalp Vector of Booking Vector
    type t_scalp_booking_vector_vector is array (natural range <>) of std_ulogic_vector;

    -- Scalp Vector of Scheduler Ack Vector
    type t_scalp_scheduler_ack_vector_vector is array (natural range <>) of std_ulogic_vector;

    type t_scalp_packet_header is record
        DstAddrxD : t_scalp_netaddr;
        SrcAddrxD : t_scalp_netaddr;
        TypexD    : integer range 0 to C_SCALP_PACKET_TYPE_RANGE_VALUE;
        LengthxD  : integer range 0 to C_SCALP_PACKET_LENGTH_RANGE_VALUE;
    end record t_scalp_packet_header;

    constant C_NO_SCALP_PACKET_HEADER : t_scalp_packet_header := (DstAddrxD => C_3D_MIN_SCALP_NETADDR,
                                                                  SrcAddrxD => C_3D_MIN_SCALP_NETADDR,
                                                                  TypexD    => 0,
                                                                  LengthxD  => 0);

    type t_scalp_packet_header_vector is array (natural range <>) of t_scalp_packet_header;

    type t_scalp_packet_word is record
        WordxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0);
        IdxD   : integer range 0 to C_SCALP_PACKET_LENGTH_RANGE_VALUE;
    end record t_scalp_packet_word;

    constant C_NO_SCALP_PACKET_WORD : t_scalp_packet_word := (WordxD => (others => '0'),
                                                              IdxD   => 0);

    type t_scalp_packet_payload is array (natural range <>) of t_scalp_packet_word;

    type t_scalp_packet_payload_vector is array (natural range <>) of t_scalp_packet_payload;

    -- constant C_NO_SCALP_PACKET_PAYLOAD : t_scalp_packet_payload := (others => C_NO_SCALP_PACKET_WORD);

    type t_scalp_packet is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload;
    end record t_scalp_packet;

    type t_scalp_packet_vector is array (natural range <>) of t_scalp_packet;

    -- constant C_NO_SCALP_PACKET : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER);

    subtype t_scalp_packet_payload_1 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_1 - 1) downto 0);
    subtype t_scalp_packet_payload_2 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_2 - 1) downto 0);
    subtype t_scalp_packet_payload_4 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_4 - 1) downto 0);
    subtype t_scalp_packet_payload_8 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_8 - 1) downto 0);
    subtype t_scalp_packet_payload_16 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_16 - 1) downto 0);
    subtype t_scalp_packet_payload_32 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_32 - 1) downto 0);
    subtype t_scalp_packet_payload_64 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_64 - 1) downto 0);
    subtype t_scalp_packet_payload_128 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_128 - 1) downto 0);
    subtype t_scalp_packet_payload_256 is t_scalp_packet_payload((C_SCALP_PACKET_PAYLOAD_SIZE_256 - 1) downto 0);

    constant C_NO_SCALP_PACKET_PAYLOAD_1   : t_scalp_packet_payload_1   := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_2   : t_scalp_packet_payload_2   := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_4   : t_scalp_packet_payload_4   := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_8   : t_scalp_packet_payload_8   := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_16  : t_scalp_packet_payload_16  := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_32  : t_scalp_packet_payload_32  := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_64  : t_scalp_packet_payload_64  := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_128 : t_scalp_packet_payload_128 := (others => C_NO_SCALP_PACKET_WORD);
    constant C_NO_SCALP_PACKET_PAYLOAD_256 : t_scalp_packet_payload_256 := (others => C_NO_SCALP_PACKET_WORD);

    type t_scalp_packet_1 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_1;
    end record t_scalp_packet_1;

    constant C_NO_SCALP_PACKET_1 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                      SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_1);

    type t_scalp_packet_2 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_2;
    end record t_scalp_packet_2;

    constant C_NO_SCALP_PACKET_2 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                      SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_2);

    type t_scalp_packet_4 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_4;
    end record t_scalp_packet_4;

    constant C_NO_SCALP_PACKET_4 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                      SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_4);

    type t_scalp_packet_8 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_8;
    end record t_scalp_packet_8;

    constant C_NO_SCALP_PACKET_8 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                      SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_8);

    type t_scalp_packet_16 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_16;
    end record t_scalp_packet_16;

    constant C_NO_SCALP_PACKET_16 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                       SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_16);

    type t_scalp_packet_32 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_32;
    end record t_scalp_packet_32;

    constant C_NO_SCALP_PACKET_32 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                       SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_32);

    type t_scalp_packet_64 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_64;
    end record t_scalp_packet_64;

    constant C_NO_SCALP_PACKET_64 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                       SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_64);

    type t_scalp_packet_128 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_128;
    end record t_scalp_packet_128;

    constant C_NO_SCALP_PACKET_128 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                        SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_128);

    type t_scalp_packet_256 is record
        SpHeaderxD  : t_scalp_packet_header;
        SpPayloadxD : t_scalp_packet_payload_256;
    end record t_scalp_packet_256;

    constant C_NO_SCALP_PACKET_256 : t_scalp_packet := (SpHeaderxD  => C_NO_SCALP_PACKET_HEADER,
                                                        SpPayloadxD => C_NO_SCALP_PACKET_PAYLOAD_256);

    function scalp_is_simulation
        return boolean;

    -- Fonction pour calculer l'index des vecteurs.
    function scalp_matrix_idx (
        IfSrcxD : integer;
        IfDstxD : integer;
        NbIfxD  : integer)
        return integer;

    -- Fonction pour changer l'endianess BIG <-> LITTLE (std_ulogic_vector)
    function change_endian_ul (
        VecxD : std_ulogic_vector)
        return std_ulogic_vector;

    -- Fonction pour changer l'endianess BIG <-> LITTLE (std_logic_vector)
    function change_endian(
        VecxD : std_logic_vector)
        return std_logic_vector;

    -- Fonction pour récupérer l'adresse de destination ou de source d'un paquet SCALP_PACKET
    function scalp_peek_addr_ul (
        DataxD : std_ulogic_vector)
        return t_scalp_netaddr;

    -- Fonction pour récupérer l'adresse de destination ou de source d'un paquet SCALP_PACKET
    function scalp_peek_addr (
        DataxD : std_logic_vector)
        return t_scalp_netaddr;

    -- Fonction pour récupérer la taille du payload.
    function scalp_peek_pld_size_ul (
        DataxD : std_ulogic_vector)
        return integer;

    -- Fonction pour récupérer la taille du payload.
    function scalp_peek_pld_size (
        DataxD : std_logic_vector)
        return integer;

    -- Fonction pour récupérer le type du paquet Scalp.
    function scalp_peek_type_ul (
        DataxD : std_ulogic_vector)
        return integer;

    -- Fonction pour récupérer le type du paquet Scalp.
    function scalp_peek_type (
        DataxD : std_logic_vector)
        return integer;

    -- Fonction de conversion Scalp Packet Header H0 vers Axis (big endian).
    function scalp_sp_header_h0_to_axis_ul (
        SpHeaderxD : t_scalp_packet_header)
        return std_ulogic_vector;

    -- Fonction de conversion Scalp Packet Header H0 vers Axis (big endian).
    function scalp_sp_header_h0_to_axis (
        SpHeaderxD : t_scalp_packet_header)
        return std_logic_vector;

    -- Fonction de conversion Scalp Packet Header H1 vers Axis (big endian).
    function scalp_sp_header_h1_to_axis_ul (
        SpHeaderxD : t_scalp_packet_header)
        return std_ulogic_vector;

    -- Fonction de conversion Scalp Packet Header H1 vers Axis (big endian).
    function scalp_sp_header_h1_to_axis (
        SpHeaderxD : t_scalp_packet_header)
        return std_logic_vector;

    -- Fonction de conversion Scalp Packet Header H2 vers Axis (big endian).
    function scalp_sp_header_h2_to_axis_ul (
        SpHeaderxD : t_scalp_packet_header)
        return std_ulogic_vector;

    -- Fonction de conversion Scalp Packet Header H2 vers Axis (big endian).
    function scalp_sp_header_h2_to_axis (
        SpHeaderxD : t_scalp_packet_header)
        return std_logic_vector;

    -- Function de conversion Scalp Packet Payload P vers Axis (big endian).
    function scalp_sp_payload_p_to_axis_ul (
        SpPayloadxD : t_scalp_packet_payload;
        IdxxD       : integer)
        return std_ulogic_vector;

    -- Function de conversion Scalp Packet Payload P vers Axis (big endian).
    function scalp_sp_payload_p_to_axis (
        SpPayloadxD : t_scalp_packet_payload;
        IdxxD       : integer)
        return std_logic_vector;

    -- Fonction de conversion Axis vers Scalp Packet Header (big endian ->
    -- little endian)
    function scalp_axis_to_sp_header_ul (
        H0xD : std_ulogic_vector;
        H1xD : std_ulogic_vector;
        H2xD : std_ulogic_vector)
        return t_scalp_packet_header;

    -- Fonction de conversion Axis vers Scalp Packet Header (big endian ->
    -- little endian)
    function scalp_axis_to_sp_header (
        H0xD : std_logic_vector;
        H1xD : std_logic_vector;
        H2xD : std_logic_vector)
        return t_scalp_packet_header;

    -- Fonction de conversion Axis vers Scalp Packet Word (big endian -> little
    -- endian)
    function scalp_axis_to_sp_word_ul (
        WordxD : std_ulogic_vector;
        IdxD   : integer range 0 to C_SCALP_PACKET_LENGTH_RANGE_VALUE)
        return t_scalp_packet_word;

    -- Fonction de conversion Axis vers Scalp Packet Word (big endian -> little
    -- endian)
    function scalp_axis_to_sp_word (
        WordxD : std_logic_vector;
        IdxD   : integer range 0 to C_SCALP_PACKET_LENGTH_RANGE_VALUE)
        return t_scalp_packet_word;

    -- Natural -> std_ulogic_vector
    function scalp_to_sulv (
        ValuexD : natural;
        SizexD  : positive)
        return std_ulogic_vector;

    -- Natural -> std_logic_vector
    function scalp_to_slv (
        ValuexD : natural;
        SizexD  : positive)
        return std_logic_vector;

    -- If Then Else -> Unsigned
    function scalp_ite (
        CondxS   : boolean;
        Value1xD : unsigned;
        Value2xD : unsigned)
        return unsigned;

    -- Fonction qui retourne le max entre deux entiers.
    function scalp_imax (
        Arg1xD : integer;
        Arg2xD : integer)
        return integer;

    -- Fonction qui retourne le log en base 2. Toujours à l'arrondu supérieur.
    function scalp_log2ceil (
        ArgxD : positive)
        return natural;

    -- Fonction qui retourne le log en base 2. Toujours à l'arrondi supérieur.
    -- La valeur de retour est >= 1.
    function scalp_log2ceilnz(
        ArgxD : positive)
        return positive;

    -- Fonction de conversion d'un code One-Hot vers binaire.
    function scalp_onehot2bin(
        OneHotxD   : std_ulogic_vector;
        EmptyValxD : integer := -1)
        return unsigned;

end package scalp_misc;

package body scalp_misc is

    -- SCALP SIMULATION
    ---------------------------------------------------------------------------
    function scalp_is_simulation
        return boolean is
        variable RetxS : boolean;
    begin  -- function scalp_is_simulation
        RetxS := false;

        if is_x('X') then
            RetxS := true;
        end if;

        return RetxS;
    end function scalp_is_simulation;

    constant C_SCALP_SIMULATION : boolean := scalp_is_simulation;
    ---------------------------------------------------------------------------

    -- Fonction pour calculer l'index des vecteurs.
    function scalp_matrix_idx (
        IfSrcxD : integer;
        IfDstxD : integer;
        NbIfxD  : integer)
        return integer is
        variable ResxD : integer := 0;
    begin
        ResxD := (IfDstxD - (IfSrcxD + 1)) mod NbIfxD;
        return ResxD;
    end function scalp_matrix_idx;

    -- Fonction pour changer l'endianess BIG <-> LITTLE (std_ulogic_vector)
    function change_endian_ul(
        VecxD : std_ulogic_vector)
        return std_ulogic_vector is
        variable VRetxD      : std_ulogic_vector(VecxD'range) := (others => '0');
        constant CNumBytesxD : natural                        := (VecxD'length / C_BYTE_SIZE);
    begin
        for i in 0 to (CNumBytesxD - 1) loop
            for j in (C_BYTE_SIZE - 1) downto 0 loop
                VRetxD((C_BYTE_SIZE * i) + j) := VecxD((C_BYTE_SIZE * (CNumBytesxD - 1 - i)) + j);
            end loop;  -- j
        end loop;  -- i

        return VRetxD;
    end function change_endian_ul;

    -- Fonction pour changer l'endianess BIG <-> LITTLE (std_logic_vector)
    function change_endian(
        VecxD : std_logic_vector)
        return std_logic_vector is
    begin
        return std_logic_vector(change_endian_ul(std_ulogic_vector(VecxD)));
    end function change_endian;

    -- Fonction pour récupérer l'adresse de destination ou de source d'un paquet SCALP_PACKET
    function scalp_peek_addr_ul (
        DataxD : std_ulogic_vector)
        return t_scalp_netaddr is
        variable AddrxD      : t_scalp_netaddr;
        variable XMapRangexD : std_ulogic_vector(0 to (C_BYTE_SIZE - 1))     := (others => '0');
        variable YMapRangexD : std_ulogic_vector(0 to (C_BYTE_SIZE - 1))     := (others => '0');
        variable ZMapRangexD : std_ulogic_vector(0 to (C_BYTE_SIZE - 1))     := (others => '0');
        variable XxD         : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0) := (others => '0');
        variable YxD         : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0) := (others => '0');
        variable ZxD         : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0) := (others => '0');
    begin
        XMapRangexD(0 to (C_BYTE_SIZE - 1)) := DataxD((DataxD'high - ((C_BYTE_SIZE * 1) - 1)) to DataxD'high);
        YMapRangexD(0 to (C_BYTE_SIZE - 1)) := DataxD((DataxD'high - ((C_BYTE_SIZE * 2) - 1)) to (DataxD'high - (C_BYTE_SIZE * 1)));
        ZMapRangexD(0 to (C_BYTE_SIZE - 1)) := DataxD((DataxD'high - ((C_BYTE_SIZE * 3) - 1)) to (DataxD'high - (C_BYTE_SIZE * 2)));

        XxD := change_endian_ul(XMapRangexD);
        YxD := change_endian_ul(YMapRangexD);
        ZxD := change_endian_ul(ZMapRangexD);

        AddrxD.XxD := to_integer(unsigned(XxD));
        AddrxD.YxD := to_integer(unsigned(YxD));
        AddrxD.ZxD := to_integer(unsigned(ZxD));

        return AddrxD;
    end function scalp_peek_addr_ul;

    -- Fonction pour récupérer l'adresse de destination ou de source d'un paquet SCALP_PACKET
    function scalp_peek_addr (
        DataxD : std_logic_vector)
        return t_scalp_netaddr is
    begin
        return scalp_peek_addr_ul(std_ulogic_vector(DataxD));
    end function scalp_peek_addr;

    -- Fonction pour récupérer la taille du payload.
    function scalp_peek_pld_size_ul (
        DataxD : std_ulogic_vector)
        return integer is
        variable PldSizexD       : std_ulogic_vector(((C_BYTE_SIZE * 2) - 1) downto 0) := (others => '0');
        variable WordBigEndianxD : std_ulogic_vector(0 to ((C_BYTE_SIZE * 2) - 1))     := (others => '0');
    begin
        WordBigEndianxD(WordBigEndianxD'left to WordBigEndianxD'right)
            := DataxD(DataxD'left to (DataxD'left + ((2 * C_BYTE_SIZE) - 1)));
        PldSizexD := change_endian_ul(WordBigEndianxD);
        return to_integer(unsigned(PldSizexD));
    end function scalp_peek_pld_size_ul;

    -- Fonction pour récupérer la taille du payload.
    function scalp_peek_pld_size (
        DataxD : std_logic_vector)
        return integer is
    begin
        return scalp_peek_pld_size_ul(std_ulogic_vector(DataxD));
    end function scalp_peek_pld_size;

    -- Fonction pour récupérer le type du paquet Scalp.
    function scalp_peek_type_ul (
        DataxD : std_ulogic_vector)
        return integer is
        variable TypexD          : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable ByteBigEndianxD : std_ulogic_vector(0 to (C_BYTE_SIZE - 1))           := (others => '0');
    begin  -- function scalp_peek_type_ul
        ByteBigEndianxD(ByteBigEndianxD'left to ByteBigEndianxD'right)
            := DataxD((DataxD'left + (C_BYTE_SIZE * 3)) to (DataxD'left + ((C_BYTE_SIZE * 4) - 1)));
        TypexD := change_endian_ul(ByteBigEndianxD);
        return to_integer(unsigned(TypexD));
    end function scalp_peek_type_ul;

    -- Fonction pour récupérer le type du paquet Scalp.
    function scalp_peek_type (
        DataxD : std_logic_vector)
        return integer is
    begin  -- function scalp_peek_type
        return scalp_peek_type_ul(std_ulogic_vector(DataxD));
    end function scalp_peek_type;

    -- Fonction pour récupérer le payload.
    function scalp_peek_pld_ul (
        DataxD : std_ulogic_vector)
        return std_ulogic_vector is
        variable PldxD                 : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable DoubleWordBigEndianxD : std_ulogic_vector(0 to ((C_BYTE_SIZE * 4) - 1))     := (others => '0');
    begin  -- function scalp_peek_pld
        DoubleWordBigEndianxD(DoubleWordBigEndianxD'left to DoubleWordBigEndianxD'right)
            := DataxD(DataxD'left to DataxD'right);
        PldxD := change_endian_ul(DoubleWordBigEndianxD);
        return PldxD;
    end function scalp_peek_pld_ul;

    -- Fonction pour récupérer le payload.
    function scalp_peek_pld (
        DataxD : std_logic_vector)
        return std_logic_vector is
    begin  -- function scalp_peek_pld
        return std_logic_vector(scalp_peek_pld_ul(std_ulogic_vector(DataxD)));
    end function scalp_peek_pld;

    -- Fonction de conversion Scalp Packet Header H0 vers Axis (big endian).
    function scalp_sp_header_h0_to_axis_ul (
        SpHeaderxD : t_scalp_packet_header)
        return std_ulogic_vector is
        variable H0xD             : std_ulogic_vector(0 to ((C_BYTE_SIZE * 4) - 1))     := (others => '0');
        variable H0LittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable DstAddrXxD       : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
        variable DstAddrYxD       : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
        variable DstAddrZxD       : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
        variable TypexD           : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
    begin
        DstAddrXxD       := std_ulogic_vector(to_unsigned(SpHeaderxD.DstAddrxD.XxD, C_BYTE_SIZE));
        DstAddrYxD       := std_ulogic_vector(to_unsigned(SpHeaderxD.DstAddrxD.YxD, C_BYTE_SIZE));
        DstAddrZxD       := std_ulogic_vector(to_unsigned(SpHeaderxD.DstAddrxD.ZxD, C_BYTE_SIZE));
        TypexD           := std_ulogic_vector(to_unsigned(SpHeaderxD.TypexD, C_BYTE_SIZE));
        H0LittleEndianxD := DstAddrXxD & DstAddrYxD & DstAddrZxD & TypexD;
        H0xD             := change_endian_ul(H0LittleEndianxD);
        return H0xD;
    end function scalp_sp_header_h0_to_axis_ul;

    -- Fonction de conversion Scalp Packet Header H0 vers Axis (big endian).
    function scalp_sp_header_h0_to_axis (
        SpHeaderxD : t_scalp_packet_header)
        return std_logic_vector is
    begin
        return std_logic_vector(scalp_sp_header_h0_to_axis_ul(SpHeaderxD));
    end function scalp_sp_header_h0_to_axis;

    -- Fonction de conversion Scalp Packet Header H1 vers Axis (big endian).
    function scalp_sp_header_h1_to_axis_ul (
        SpHeaderxD : t_scalp_packet_header)
        return std_ulogic_vector is
        variable H1xD             : std_ulogic_vector(0 to ((C_BYTE_SIZE * 4) - 1))     := (others => '0');
        variable H1LittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable SrcAddrXxD       : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
        variable SrcAddrYxD       : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
        variable SrcAddrZxD       : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
        variable PadxD            : std_ulogic_vector((C_BYTE_SIZE - 1) downto 0)       := (others => '0');
    begin
        SrcAddrXxD       := std_ulogic_vector(to_unsigned(SpHeaderxD.SrcAddrxD.XxD, C_BYTE_SIZE));
        SrcAddrYxD       := std_ulogic_vector(to_unsigned(SpHeaderxD.SrcAddrxD.YxD, C_BYTE_SIZE));
        SrcAddrZxD       := std_ulogic_vector(to_unsigned(SpHeaderxD.SrcAddrxD.ZxD, C_BYTE_SIZE));
        -- 31 downto 0
        H1LittleEndianxD := SrcAddrXxD & SrcAddrYxD & SrcAddrZxD & PadxD;
        H1xD             := change_endian_ul(H1LittleEndianxD);
        return H1xD;
    end function scalp_sp_header_h1_to_axis_ul;

    -- Fonction de conversion Scalp Packet Header H1 vers Axis (big endian).
    function scalp_sp_header_h1_to_axis (
        SpHeaderxD : t_scalp_packet_header)
        return std_logic_vector is
    begin
        return std_logic_vector(scalp_sp_header_h1_to_axis_ul(SpHeaderxD));
    end function scalp_sp_header_h1_to_axis;

    -- Fonction de conversion Scalp Packet Header H2 vers Axis (big endian).
    function scalp_sp_header_h2_to_axis_ul (
        SpHeaderxD : t_scalp_packet_header)
        return std_ulogic_vector is
        variable H2xD             : std_ulogic_vector(0 to ((C_BYTE_SIZE * 4) - 1))     := (others => '0');
        variable H2LittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable PldSizexD        : std_ulogic_vector(((C_BYTE_SIZE * 2) - 1) downto 0) := (others => '0');
        variable PadxD            : std_ulogic_vector(((C_BYTE_SIZE * 2) - 1) downto 0) := (others => '0');
    begin
        PldSizexD        := std_ulogic_vector(to_unsigned(SpHeaderxD.LengthxD, (C_BYTE_SIZE * 2)));
        H2LittleEndianxD := PldSizexD & PadxD;
        H2xD             := change_endian_ul(H2LittleEndianxD);
        return H2xD;
    end function scalp_sp_header_h2_to_axis_ul;

    -- Fonction de conversion Scalp Packet Header H2 vers Axis (big endian).
    function scalp_sp_header_h2_to_axis (
        SpHeaderxD : t_scalp_packet_header)
        return std_logic_vector is
    begin
        return std_logic_vector(scalp_sp_header_h2_to_axis_ul(SpHeaderxD));
    end function scalp_sp_header_h2_to_axis;

    -- Fonction de conversion Scalp Packet Payload P vers Axis (big endian).
    function scalp_sp_payload_p_to_axis_ul (
        SpPayloadxD : t_scalp_packet_payload;
        IdxxD       : integer)
        return std_ulogic_vector is
        variable PxD             : std_ulogic_vector(0 to ((C_BYTE_SIZE * 4) - 1))     := (others => '0');
        variable PLittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
    begin
        PLittleEndianxD(PLittleEndianxD'left downto PLittleEndianxD'right) :=
            SpPayloadxD(IdxxD).WordxD(SpPayloadxD(IdxxD).WordxD'left downto SpPayloadxD(IdxxD).WordxD'right);
        PxD := change_endian_ul(PLittleEndianxD);
        return PxD;
    end function scalp_sp_payload_p_to_axis_ul;

    -- Fonction de conversion Scalp Packet Payload P vers Axis (big endian).
    function scalp_sp_payload_p_to_axis (
        SpPayloadxD : t_scalp_packet_payload;
        IdxxD       : integer)
        return std_logic_vector is
    begin
        return std_logic_vector(scalp_sp_payload_p_to_axis_ul(SpPayloadxD, IdxxD));
    end function scalp_sp_payload_p_to_axis;

    -- Fonction de conversion Axis vers Scalp Packet Header (big endian ->
    -- little endian)
    function scalp_axis_to_sp_header_ul (
        H0xD : std_ulogic_vector;
        H1xD : std_ulogic_vector;
        H2xD : std_ulogic_vector)
        return t_scalp_packet_header is
        variable H0LittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable H1LittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable H2LittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable DstAddrXxD       : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable DstAddrYxD       : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable DstAddrZxD       : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable SrcAddrXxD       : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable SrcAddrYxD       : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable SrcAddrZxD       : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable TypexD           : std_ulogic_vector(((C_BYTE_SIZE * 1) - 1) downto 0) := (others => '0');
        variable LengthxD         : std_ulogic_vector(((C_BYTE_SIZE * 2) - 1) downto 0) := (others => '0');
        variable SpHeaderxD       : t_scalp_packet_header                               := C_NO_SCALP_PACKET_HEADER;
    begin  -- function scalp_axis_to_sp_header_ul
        -- Big endian to little endian
        H0LittleEndianxD := change_endian_ul(H0xD);
        H1LittleEndianxD := change_endian_ul(H1xD);
        H2LittleEndianxD := change_endian_ul(H2xD);

        -- Dst Addr + Type (H0)
        DstAddrXxD := H0LittleEndianxD(H0LittleEndianxD'high downto (H0LittleEndianxD'high - ((C_BYTE_SIZE * 1) - 1)));
        DstAddrYxD := H0LittleEndianxD((H0LittleEndianxD'high - (C_BYTE_SIZE * 1)) downto
                                       (H0LittleEndianxD'high - ((C_BYTE_SIZE * 2) - 1)));
        DstAddrZxD := H0LittleEndianxD((H0LittleEndianxD'high - (C_BYTE_SIZE * 2)) downto
                                       (H0LittleEndianxD'high - ((C_BYTE_SIZE * 3) - 1)));
        TypexD := H0LittleEndianxD((H0LittleEndianxD'high - (C_BYTE_SIZE * 3)) downto
                                   (H0LittleEndianxD'high - ((C_BYTE_SIZE * 4) - 1)));
        -- Src Addr (H1)
        SrcAddrXxD := H1LittleEndianxD(H1LittleEndianxD'high downto (H1LittleEndianxD'high - ((C_BYTE_SIZE * 1) - 1)));
        SrcAddrYxD := H1LittleEndianxD((H1LittleEndianxD'high - (C_BYTE_SIZE * 1)) downto
                                       (H1LittleEndianxD'high - ((C_BYTE_SIZE * 2) - 1)));
        SrcAddrZxD := H1LittleEndianxD((H1LittleEndianxD'high - (C_BYTE_SIZE * 2)) downto
                                       (H1LittleEndianxD'high - ((C_BYTE_SIZE * 3) - 1)));
        -- Payload Size (H2)
        LengthxD := H2LittleEndianxD(H2LittleEndianxD'high downto (H2LittleEndianxD'high - ((C_BYTE_SIZE * 2) - 1)));

        -- Fill Scalp Packet Header
        SpHeaderxD.DstAddrxD.XxD := to_integer(unsigned(DstAddrXxD));
        SpHeaderxD.DstAddrxD.YxD := to_integer(unsigned(DstAddrYxD));
        SpHeaderxD.DstAddrxD.ZxD := to_integer(unsigned(DstAddrZxD));
        SpHeaderxD.TypexD        := to_integer(unsigned(TypexD));
        SpHeaderxD.SrcAddrxD.XxD := to_integer(unsigned(SrcAddrXxD));
        SpHeaderxD.SrcAddrxD.YxD := to_integer(unsigned(SrcAddrYxD));
        SpHeaderxD.SrcAddrxD.ZxD := to_integer(unsigned(SrcAddrZxD));
        SpHeaderxD.LengthxD      := to_integer(unsigned(LengthxD));

        return SpHeaderxD;
    end function scalp_axis_to_sp_header_ul;

    -- Fonction de conversion Axis vers Scalp Packet Header (big endian ->
    -- little endian)
    function scalp_axis_to_sp_header (
        H0xD : std_logic_vector;
        H1xD : std_logic_vector;
        H2xD : std_logic_vector)
        return t_scalp_packet_header is
    begin  -- function scalp_axis_to_sp_header
        return scalp_axis_to_sp_header_ul(std_ulogic_vector(H0xD),
                                          std_ulogic_vector(H1xD),
                                          std_ulogic_vector(H2xD));
    end function scalp_axis_to_sp_header;

    -- Fonction de conversion Axis vers Scalp Packet Word (big endian -> little
    -- endian)
    function scalp_axis_to_sp_word_ul (
        WordxD : std_ulogic_vector;
        IdxD   : integer range 0 to C_SCALP_PACKET_LENGTH_RANGE_VALUE)
        return t_scalp_packet_word is
        variable WordLittleEndianxD : std_ulogic_vector(((C_BYTE_SIZE * 4) - 1) downto 0) := (others => '0');
        variable SpWordxD           : t_scalp_packet_word                                 := C_NO_SCALP_PACKET_WORD;
    begin  -- function scalp_axis_to_sp_word_ul
        WordLittleEndianxD := change_endian_ul(WordxD);
        SpWordxD.WordxD    := WordLittleEndianxD;
        SpWordxD.IdxD      := IdxD;
        return SpWordxD;
    end function scalp_axis_to_sp_word_ul;

    -- Fonction de conversion Axis vers Scalp Packet Word (big endian -> little
    -- endian)
    function scalp_axis_to_sp_word (
        WordxD : std_logic_vector;
        IdxD   : integer range 0 to C_SCALP_PACKET_LENGTH_RANGE_VALUE)
        return t_scalp_packet_word is
    begin  -- function scalp_axis_to_sp_word
        return scalp_axis_to_sp_word_ul(std_ulogic_vector(WordxD), IdxD);
    end function scalp_axis_to_sp_word;

    -- Natural -> std_ulogic_vector
    function scalp_to_sulv (
        ValuexD : natural;
        SizexD  : positive)
        return std_ulogic_vector is
        constant ResxD : std_ulogic_vector((SizexD - 1) downto 0) := std_ulogic_vector(to_unsigned(ValuexD, SizexD));
    begin  -- function scalp_to_sulv
        return ResxD;
    end function scalp_to_sulv;

    -- Natural -> std_logic_vector
    function scalp_to_slv (
        ValuexD : natural;
        SizexD  : positive)
        return std_logic_vector is
        constant ResxD : std_logic_vector((SizexD - 1) downto 0) := std_logic_vector(to_unsigned(ValuexD, SizexD));
    begin  -- function scalp_to_slv
        return ResxD;
    end function scalp_to_slv;

    -- If Then Else -> Unsigned
    function scalp_ite (
        CondxS   : boolean;
        Value1xD : unsigned;
        Value2xD : unsigned)
        return unsigned is
    begin  -- function scalp_ite
        if CondxS then
            return Value1xD;
        else
            return Value2xD;
        end if;
    end function scalp_ite;

    -- Fonction qui retourne le max entre deux entiers.
    function scalp_imax (
        Arg1xD : integer;
        Arg2xD : integer)
        return integer is
    begin  -- function scalp_imax
        if Arg1xD > Arg2xD then
            return Arg1xD;
        end if;

        return Arg2xD;
    end function scalp_imax;

    -- Fonction qui retourne le log en base 2. Toujours à l'arrondu supérieur.
    function scalp_log2ceil (
        ArgxD : positive)
        return natural is
        variable TmpxD : positive;
        variable LogxD : natural;
    begin  -- function scalp_log2ceil
        if ArgxD = 1 then
            return 0;
        end if;

        TmpxD := 1;
        LogxD := 0;

        while ArgxD > TmpxD loop
            TmpxD := TmpxD * 2;
            LogxD := LogxD + 1;
        end loop;

        return LogxD;
    end function scalp_log2ceil;

    -- Fonction qui retourne le log en base 2. Toujours à l'arrondi supérieur.
    -- La valeur de retour est >= 1.
    function scalp_log2ceilnz (
        ArgxD : positive)
        return positive is
    begin  -- function scalp_log2ceilnz
        return scalp_imax(1, scalp_log2ceil(ArgxD));
    end function scalp_log2ceilnz;

    -- Fonction de conversion d'un code One-Hot vers binaire.
    function scalp_onehot2bin(
        OneHotxD   : std_ulogic_vector;
        EmptyValxD : integer := -1)
        return unsigned is
        variable ResxD : unsigned((scalp_log2ceilnz(scalp_imax(OneHotxD'high, EmptyValxD) + 1) - 1) downto 0);
        variable ChkxD : natural;
    begin  -- function scalp_onehot2bin
        if (EmptyValxD > 0) and (OneHotxD = (OneHotxD'range => '0')) then
            ResxD := to_unsigned(EmptyValxD, ResxD'length);
        else
            ResxD := (others => '0');
            ChkxD := 0;

            for i in OneHotxD'range loop
                if OneHotxD(i) = '1' then
                    ResxD := ResxD or to_unsigned(i, ResxD'length);
                    ChkxD := ChkxD + 1;
                end if;
            end loop;  -- i
        end if;

        if C_SCALP_SIMULATION and (ChkxD /= 1) and ((ChkxD > 1) or (EmptyValxD < 0)) then
            report "Broken 1-Hot-Code with " & integer'image(ChkxD) & " bits set." severity warning;
            ResxD := (others => 'X');
        end if;
        return ResxD;
    end function scalp_onehot2bin;

    ---------------------------------------------------------------------------
    -- Fonction Random
    ---------------------------------------------------------------------------
    -- Random Real Value
    -- impure function scalp_rand_real (
    --     MinValxD : real;
    --     MaxValxD : real)
    --     return real is
    --     variable RandxD : real;
    -- begin  -- function scalp_rand_real
    --     uniform(Seed1xD, Seed2xD, RandxD);
    --     return RandxD * (MaxValxD - MinValxD) + MinValxD;
    -- end function scalp_rand_real;

    -- Random Integer Value
    -- impure function scalp_rand_int (
    --     MinValxD : integer;
    --     MaxValxD : integer)
    --     return integer is
    --     variable RandxD : real;
    -- begin  -- function scalp_rand_int
    --     uniform(Seed1xD, Seed2xD, RandxD);
    --     return integer(round(RandxD * real(MaxValxD - MinValxD + 1) + real(MinValxD) - 0.5));
    -- end function scalp_rand_int;

    -- Random std_ulogic_vector
    -- impure function scalp_rand_sulv (
    --     LenxD : integer)
    --     return std_ulogic_vector is
    --     variable RandxD : real;
    --     variable SlvxD  : std_ulogic_vector((LenxD - 1) downto 0);
    -- begin  -- function scalp_rand_slv
    --     for i in SlvxD'range loop
    --         uniform(Seed1xD, Seed2xD, RandxD);
    --         SlvxD(i) := '1' when RandxD > 0.5 else '0';
    --     end loop;  -- i

    --     return SlvxD;
    -- end function scalp_rand_sulv;

    -- Random Time
    -- impure function scalp_rand_time (
    --     MinValxD : time;
    --     MaxValxD : time;
    --     UnitxD   : time := ns)
    --     return time is
    --     variable RandxD       : real;
    --     variable RandScaledxD : real;
    --     variable MinRealxD    : real;
    --     variable MaxRealxD    : real;
    -- begin  -- function scalp_rand_time
    --     uniform(Seed1xD, Seed2xD, RandxD);
    --     MinRealxD    := real(MinValxD / UnitxD);
    --     MaxRealxD    := real(MaxValxD / UnitxD);
    --     RandScaledxD := RandxD * (MaxRealxD - MinRealxD) + MinRealxD;
    --     return real(RandScaledxD) * UnitxD;
    -- end function scalp_rand_time;

end package body scalp_misc;
