----------------------------------------------------------------------------------
--                                 _             _
--                                | |_  ___ _ __(_)__ _
--                                | ' \/ -_) '_ \ / _` |
--                                |_||_\___| .__/_\__,_|
--                                         |_|
--
----------------------------------------------------------------------------------
--
-- Company: hepia
-- Author: Joachim Schmidt <joachim.schmidt@hesge.ch>
--
-- Module Name: scalp_tx_side.vhd
-- Target Device: SCALP xc7z015clg485-2
-- Tool version: 2019.1
-- Description: Scalp TX side state machine.
--
-- Last update: 2020-11-30
--
---------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

--use ieee.std_logic_unsigned.all;
--use ieee.std_logic_arith.all;
--use ieee.std_logic_misc.all;

--library UNISIM;
--use UNISIM.VCOMPONENTS.all;

use work.axi4_pkg.all;
use work.scalp_misc.all;

entity scalp_tx_side is

    generic (
        C_SCALP_SCHEDULER_ACK_VECTOR_SIZE : integer range 0 to 255 := 1);

    port (
        -- System Clock and Reset
        SysClkxCI             : in  std_ulogic;
        SysRstxRNAI           : in  std_ulogic;
        -- Routeur Network Address
        DstRouterNetAddrxDO   : out t_scalp_netaddr;
        RouterNetAddrValidxSO : out std_ulogic;
        -- Indicates when the booking vector is valid.
        BookingVectorValidxSI : in  std_ulogic;
        -- Scheduler Acknowledgement Vector
        -- SchedulerAckxDI       : in  std_ulogic_vector((C_SCALP_SCHEDULER_ACK_VECTOR_SIZE - 1) downto 0);
        -- Axi4 Stream
        -- Input side
        RXAxi4M2SLinkxDI      : in  t_axi4m2s;
        RXAxi4S2MLinkxDO      : out t_axi4s2m;
        -- Output Side
        TXAxi4M2SLinkxDO      : out t_axi4m2s;
        TXAxi4S2MLinkxDI      : in  t_axi4s2m);

end entity scalp_tx_side;

architecture behavioral of scalp_tx_side is

    component scalp_fifo_double_register is
        port (
            -- System Clock and Reset
            SysClkxCI      : in  std_ulogic;
            SysRstxRNAI    : in  std_ulogic;
            -- Axis Ports (Slave Side)
            Axism2sLinkxDI : in  t_axi4m2s;
            Axiss2mLinkxDO : out t_axi4s2m;
            -- Axis Ports (Master Side)
            Axism2sLinkxDO : out t_axi4m2s;
            Axiss2mLinkxDI : in  t_axi4s2m);
    end component scalp_fifo_double_register;

    signal DstRouterNetAddrxDP   : t_scalp_netaddr        := C_3D_MIN_SCALP_NETADDR;
    signal DstRouterNetAddrxDN   : t_scalp_netaddr        := C_3D_MIN_SCALP_NETADDR;
    signal RouterNetAddrValidxSP : std_ulogic             := '0';
    signal RouterNetAddrValidxSN : std_ulogic             := '0';
    signal BookingVectorValidxS  : std_ulogic             := '0';
    -- signal SchedulerAckxD        : std_ulogic_vector((C_SCALP_SCHEDULER_ACK_VECTOR_SIZE - 1) downto 0) := (others => '0');
    signal RXAxi4M2SLinkxD       : t_axi4m2s              := C_NO_AXI4_M2S;
    signal RXAxi4S2MLinkxD       : t_axi4s2m              := C_NO_AXI4_S2M;
    signal TXAxi4M2SLinkxD       : t_axi4m2s              := C_NO_AXI4_M2S;
    signal TXAxi4S2MLinkxD       : t_axi4s2m              := C_NO_AXI4_S2M;
    -- TX Side States
    signal TXSideStatexDP        : t_scalp_tx_side_states := E_SCALP_TX_IDLE;
    signal TXSideStatexDN        : t_scalp_tx_side_states := E_SCALP_TX_IDLE;
    -- Others
    signal ReadyInxD             : t_axi4s2m              := C_NO_AXI4_S2M;
    signal EnPopNPushxS          : std_ulogic             := '0';

begin  -- architecture behavioral

    -- Asynchronous Statements

    EntityIOxB : block is
    begin  -- block EntityIOxB
        DstRouterNetAddrxAS   : DstRouterNetAddrxDO   <= DstRouterNetAddrxDP;
        -- SchedulerAckxAS       : SchedulerAckxD        <= SchedulerAckxDI;
        RXAxi4M2SLinkxAS      : RXAxi4M2SLinkxD       <= RXAxi4M2SLinkxDI;
        RXAxi4S2MLinkxAS      : RXAxi4S2MLinkxDO      <= RXAxi4S2MLinkxD;
        TXAxi4M2SLinkxAS      : TXAxi4M2SLinkxDO      <= TXAxi4M2SLinkxD;
        TXAxi4S2MLinkxAS      : TXAxi4S2MLinkxD       <= TXAxi4S2MLinkxDI;
        RouterNetAddrValidxAS : RouterNetAddrValidxSO <= RouterNetAddrValidxSP;
        BookingVectorValidxAS : BookingVectorValidxS  <= BookingVectorValidxSI;
    end block EntityIOxB;

    AsyncStatementsxB : block is
    begin  -- block AsyncStatementsxB
        ReadyInxAS : ReadyInxD <= TXAxi4S2MLinkxD when EnPopNPushxS = '1' else C_NO_AXI4_S2M;
    end block AsyncStatementsxB;

    -- Synchronous Statements

    ScalpFifoDoubleRegisterxI : entity work.scalp_fifo_double_register
        port map (
            -- System Clock and Reset
            SysClkxCI      => SysClkxCI,
            SysRstxRNAI    => SysRstxRNAI,
            -- Axis Ports (Slave Side)
            Axism2sLinkxDI => RXAxi4M2SLinkxD,
            Axiss2mLinkxDO => RXAxi4S2MLinkxD,
            -- Axis Ports (Master Side)
            Axism2sLinkxDO => TXAxi4M2SLinkxD,
            Axiss2mLinkxDI => ReadyInxD);

    UpdateRegxP : process (SysClkxCI, SysRstxRNAI) is
    begin  -- process UpdateRegxP
        if SysRstxRNAI = '0' then
            TXSideStatexDP        <= E_SCALP_TX_IDLE;
            DstRouterNetAddrxDP   <= C_3D_MIN_SCALP_NETADDR;
            RouterNetAddrValidxSP <= '0';
        elsif rising_edge(SysClkxCI) then
            TXSideStatexDP        <= TXSideStatexDN;
            DstRouterNetAddrxDP   <= DstRouterNetAddrxDN;
            RouterNetAddrValidxSP <= RouterNetAddrValidxSN;
        end if;
    end process UpdateRegxP;

    TXSideStateMachinexP : process (BookingVectorValidxS, DstRouterNetAddrxDP,
                                    RXAxi4M2SLinkxD.DataxD,
                                    RXAxi4M2SLinkxD.ValidxS,
                                    TXAxi4M2SLinkxD.LastxS, TXSideStatexDP) is
    begin  -- process TXSideStateMachinexP
        -- Default values
        TXSideStatexDN        <= TXSideStatexDP;
        DstRouterNetAddrxDN   <= DstRouterNetAddrxDP;
        RouterNetAddrValidxSN <= '0';
        EnPopNPushxS          <= '0';

        case TXSideStatexDP is
            when E_SCALP_TX_IDLE =>
                    -- Si un paquet scalp est présent, on lit le premier mot
                    -- pour en extraire l'adresse de destination.
                    if RXAxi4M2SLinkxD.ValidxS = '1' then
                        DstRouterNetAddrxDN   <= scalp_peek_addr_ul(RXAxi4M2SLinkxD.DataxD);
                        RouterNetAddrValidxSN <= '1';
                        TXSideStatexDN        <= E_SCALP_TX_WAIT_BOOKING_VECTOR;
                    end if;

            when E_SCALP_TX_WAIT_BOOKING_VECTOR =>
                    if BookingVectorValidxS = '1' then
                        EnPopNPushxS   <= '1';
                        TXSideStatexDN <= E_SCALP_TX_POP_N_PUSH;
                    end if;

            when E_SCALP_TX_POP_N_PUSH =>
                    EnPopNPushxS <= '1';

                    if TXAxi4M2SLinkxD.LastxS = '1' then
                        TXSideStatexDN <= E_SCALP_TX_IDLE;
                    end if;

            when others => null;
        end case;
    end process TXSideStateMachinexP;

end architecture behavioral;
